/*
Szymon Rusinkiewicz
Princeton University

xshade-main.cc
Exaggerated shading: local-light multiscale detail visualization
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "xsh.h"
#include "shade.h"
#include "colormap.h"
#include "XForm.h"
#include "GLCamera.h"
#include "GL/glui.h"


// Globals: mesh and camera
TriMesh *mesh;
XSHdata *data;
GLCamera camera;
float fov = 0.7f;
xform xf;
const char *filename;
char *xffilename;


// Options
float base_contrast, detail_contrast, curv_contrast;
float log_base_thresh, log_detail_thresh, log_curv_thresh;
float detail_freqdep, curv_freqdep;
int only_band = -1;
int skip_bands = 0;
int plain_cosine = 0;
int use_pdir = 0;
point attention;
float attention_r = 0.0f;
float light_azim = 135, light_azim2_rel = -70, light_altitude = 0;


// Signal a redraw
void need_redraw()
{
	glutPostRedisplay();
}


// Compute multipliers and thresholds for the different scales
void compute_mults(const XSHdata *data,
		   float contrast, float log_thresh, float freqdep,
		   vector<float> &mult, vector<float> &thresh)
{
	int n = data->sigmas.size() - 1;
	mult.clear();  mult.resize(n);
	thresh.clear();  thresh.resize(n);

	float sum = 0.0f;
	for (int i = 0; i < n; i++) {
		if (i < skip_bands || only_band == 0 ||
		    only_band > 0 && only_band != i+1)
			continue;
		mult[i] = pow(data->sigmas[i], -freqdep);
		sum += mult[i];
	}
	if (sum) {
		for (int i = 0; i < n; i++) {
			mult[i] *= contrast / sum;
			thresh[i] = mult[i] * pow(2.0f, log_thresh);
		}
	}
}


// Compute texture coordinates...  Actually calls the shading functions
void compute_texcoords(vector<float> &texcoords)
{
	int nv = mesh->vertices.size();
	texcoords.resize(2*nv);

	// Find light directions
	float th = M_PI/180.0f * light_altitude;
	float ph = M_PI/180.0f * light_azim;
	vec lightdir(cos(th)*cos(ph), cos(th)*sin(ph), sin(th));
	lightdir = inv(norm_xf(xf)) * lightdir;

	ph += M_PI/180.0f * light_azim2_rel;
	vec lightdir2(cos(th)*cos(ph), cos(th)*sin(ph), sin(th));
	lightdir2 = inv(norm_xf(xf)) * lightdir2;

	// Just plain cosine
	if (plain_cosine) {
		shade_cosine(data->snormals[0], lightdir,
			1.0f, 1.0f, texcoords, 0);
		if (colormap >= COLORMAP_RAINBOW)
			shade_cosine(data->snormals[0], lightdir2,
				1.0f, 1.0f, texcoords, 1);
		return;
	}

	// Multi-pass shading.  Start with the base coat...
	float bc = base_contrast;
	if (only_band > 0)
		bc = 0.0f;
	shade_cosine(data->snormals.back(), lightdir,
		bc, pow(2.0f, log_base_thresh),
		texcoords, 0);
	if (colormap >= COLORMAP_RAINBOW)
		shade_cosine(data->snormals.back(), lightdir2,
			bc, pow(2.0f, log_base_thresh),
			texcoords, 1);

	// Now do exaggerated shading
	static vector<float> mult, thresh;
	if (detail_contrast > 0.0f) {
		compute_mults(data, 100.0f * detail_contrast,
			log_detail_thresh, detail_freqdep, mult, thresh);
		shade_exaggerated(mesh, data, lightdir,
			use_pdir, mult, thresh,
			attention, attention_r, texcoords, 0);
		if (colormap >= COLORMAP_RAINBOW)
			shade_exaggerated(mesh, data, lightdir2,
				use_pdir, mult, thresh,
				attention, attention_r, texcoords, 1);
	}

	// And mean curvature shading
	if (curv_contrast > 0.0f) {
		compute_mults(data, 10.0f * curv_contrast,
			log_curv_thresh, curv_freqdep, mult, thresh);
		shade_meancurv(mesh, data, mult, thresh,
			attention, attention_r, texcoords, 0);
		if (colormap >= COLORMAP_RAINBOW)
			shade_meancurv(mesh, data, mult, thresh,
				attention, attention_r, texcoords, 1);
	}
}


// Clear the screen and reset OpenGL modes to something sane
void cls()
{
	glDisable(GL_DITHER);
	glDisable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_NORMALIZE);
	glDisable(GL_LIGHTING);
	glDisable(GL_COLOR_MATERIAL);
	set_bgcolor();
	glClearDepth(1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}


// Draw triangle strips.  They are stored as length followed by values.
void draw_tstrips()
{
	const int *t = &mesh->tstrips[0];
	const int *end = t + mesh->tstrips.size();
	while (likely(t < end)) {
		int striplen = *t++;
		glDrawElements(GL_TRIANGLE_STRIP, striplen, GL_UNSIGNED_INT, t);
		t += striplen;
	}
}


// Draw the scene
void redraw()
{
	timestamp t = now();
	GLUI_Master.auto_set_viewport();
	camera.setupGL(xf * mesh->bsphere.center, mesh->bsphere.r);
	cls();

	glPushMatrix();
	glMultMatrixd((double *)xf);

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);

	static vector<float> texcoords;
	compute_texcoords(texcoords);

	setuptexture();
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, &mesh->vertices[0][0]);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer(2, GL_FLOAT, 0, &texcoords[0]);

	draw_tstrips();

	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);

	glPopMatrix();
	glutSwapBuffers();
	if (camera.autospin(xf))
		need_redraw();

	printf("\r                                         \r");
	printf("%.1f msec.\r", 1000.0f * (now() - t));
	fflush(stdout);
	GLUI_Master.sync_live_all();
}


// Auto-adjust the contrast for full range
void auto_detail_contrast(int dummy = 0)
{
	if (plain_cosine)
		return;

	int nv = mesh->vertices.size();
	vector<float> texcoords(2*nv);
	float last_contrast = 0.0f;
	for (int pass = 0; pass < 5; pass++) {
		if (last_contrast / detail_contrast < 1.01f &&
		    last_contrast / detail_contrast > 0.99f)
			break;
		compute_texcoords(texcoords);
		for (int i = 0; i < nv; i++)
			texcoords[i] = texcoords[2*i];
		const float frac = 0.01f;
		int which = int(frac * nv);
		nth_element(texcoords.begin(),
			    texcoords.begin() + which,
			    texcoords.begin() + nv);
		float val = fabs(texcoords[which]);
		which = int((1.0f - frac) * nv);
		nth_element(texcoords.begin(),
			    texcoords.begin() + which,
			    texcoords.begin() + nv);
		val = 0.5f * (val + fabs(texcoords[which]));

		last_contrast = detail_contrast;
		float newcontrast = detail_contrast / val;
		printf("newcontrast = %f\n", newcontrast);
		if (newcontrast > 0.01f && newcontrast < 2.0f)
			detail_contrast = newcontrast;
	}
}


// Save the current image to a PPM file
void dump_image(int dummy = 0)
{
	// Find first non-used filename
	char *basename = new char[strlen(filename) + 1];
	strcpy(basename, filename);
	char *dot = strrchr(basename, '.');
	if (dot)
		*dot = '\0';

	const char filenamepattern[] = "%s_%d.ppm";
	int imgnum = 0;
	FILE *f;
	while (1) {
		char imfilename[1024];
		sprintf(imfilename, filenamepattern, basename, imgnum++);
		f = fopen(imfilename, "rb");
		if (!f) {
			f = fopen(imfilename, "wb");
			printf("\n\nSaving image %s... ", imfilename);
			fflush(stdout);
			break;
		}
		fclose(f);
	}

	// Read pixels
	GLUI_Master.auto_set_viewport();
	int V[4];
	glGetIntegerv(GL_VIEWPORT, V);
	int width = V[2], height = V[3];
	char *buf = new char[width*height*3];
	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	glReadPixels(V[0], V[1], width, height, GL_RGB, GL_UNSIGNED_BYTE, buf);

	// Flip top-to-bottom
	for (int i = 0; i < height/2; i++) {
		char *row1 = buf + 3 * width * i;
		char *row2 = buf + 3 * width * (height - 1 - i);
		for (int j = 0; j < 3 * width; j++)
			swap(row1[j], row2[j]);
	}

	// Write out file
	fprintf(f, "P6\n%d %d\n255\n", width, height);
	fwrite(buf, width*height*3, 1, f);
	fclose(f);
	delete [] buf;

	printf("Done.\n\n");
}


// Load up preset parameters
void preset(int p)
{
	plain_cosine = 0;
	if (p == 2) {
		base_contrast = 1.0;
		detail_contrast = 0.25;
		curv_contrast = 0.1;
		log_base_thresh = 0.0;
		log_detail_thresh = -4.5;
		log_curv_thresh = -4.5;
		detail_freqdep = 0;
		curv_freqdep = 0;
		use_pdir = 1;
		skip_bands = 0;
	} else if (p == 3) {
		base_contrast = 0.7;
		detail_contrast = 0.2;
		curv_contrast = 0.2;
		log_base_thresh = 0.0;
		log_detail_thresh = -3.5;
		log_curv_thresh = -4.5;
		detail_freqdep = 0.9;
		curv_freqdep = 0.9;
		use_pdir = 1;
		skip_bands = 0;
	} else {
		base_contrast = 0.8;
		detail_contrast = 0.2;
		curv_contrast = 0.0;
		log_base_thresh = 0.0;
		log_detail_thresh = -4.5;
		log_curv_thresh = -4.5;
		detail_freqdep = 0.5;
		curv_freqdep = 0.5;
		use_pdir = 1;
		skip_bands = 0;
	}

	// Just reset the lighting for presets...
	light_azim = 135;
	light_azim2_rel = -70;
	light_altitude = 0;

	need_redraw();
	GLUI_Master.sync_live_all();
}


// Set the view to look at the middle of the mesh, from reasonably far away
void resetview()
{
	camera.stopspin();

	if (!xf.read(xffilename))
		xf = xform::trans(0, 0, -3.5f / fov * mesh->bsphere.r) *
		     xform::trans(-mesh->bsphere.center);
}


// Set the attention point
void set_attention(int x, int y, bool click)
{
	// Read depth at clicked-on pixel
	GLdouble M[16], P[16];
	GLint V[4];
	M[0] = M[5] = M[10] = M[15] = 1.0;
	M[1] = M[2] = M[3] = M[4] =
	M[6] = M[7] = M[8] = M[9] =
	M[11] = M[12] = M[13] = M[14] = 0;
	glGetDoublev(GL_PROJECTION_MATRIX, P);
	glGetIntegerv(GL_VIEWPORT, V);

	y = V[1] + V[3] - 1 - y;
	float d;
	glReadPixels(x, y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &d);

	if (d < 0.0001f || d > 0.9999f) {
		if (click)
			attention_r = 0.0f;
		return;
	}

	GLdouble X, Y, Z;
	gluUnProject(x, y, d, M, P, V, &X, &Y, &Z);
	point p = inv(xf) * point(X, Y, Z);

	if (click) {
		attention = p;
		attention_r = data->sigmas[0];
	} else {
		attention_r = dist(attention, p);
	}
}


// Handle mouse button and motion events
static unsigned buttonstate = 0;
static const unsigned ctrl_pressed = 1 << 30;

void mousemotionfunc(int x, int y)
{
	if (buttonstate & ctrl_pressed) {
		set_attention(x, y, false);
		need_redraw();
		return;
	}

	static const Mouse::button physical_to_logical_map[] = {
		Mouse::NONE, Mouse::ROTATE, Mouse::MOVEXY, Mouse::MOVEZ,
		Mouse::MOVEZ, Mouse::MOVEXY, Mouse::MOVEXY, Mouse::MOVEXY,
	};
	Mouse::button b = Mouse::NONE;
	if (buttonstate & (1 << 3))
		b = Mouse::WHEELUP;
	else if (buttonstate & (1 << 4))
		b = Mouse::WHEELDOWN;
	else
		b = physical_to_logical_map[buttonstate & 7];

	camera.mouse(x, y, b,
		     xf * mesh->bsphere.center,
		     mesh->bsphere.r, xf);

	need_redraw();
	GLUI_Master.sync_live_all();
}

void mousebuttonfunc(int button, int state, int x, int y)
{
	if (state == GLUT_DOWN)
		buttonstate |= (1 << button);
	else
		buttonstate &= ~(1 << button);

	if (glutGetModifiers() & GLUT_ACTIVE_CTRL)
		buttonstate |= ctrl_pressed;
	else
		buttonstate &= ~ctrl_pressed;

	if (buttonstate & ctrl_pressed) {
		if (state == GLUT_DOWN)
			set_attention(x, y, true);
		else if (attention_r == data->sigmas[0])
			attention_r = 0.0f;
		need_redraw();
		return;
	}

	mousemotionfunc(x, y);
}


#define Ctrl (1-'a')
// Keyboard callback
void keyboardfunc(unsigned char key, int x, int y)
{
	switch (key) {
		case ' ':
			resetview();
			break;
		case 'I':
			dump_image();
			break;
		case 'l':
			colormap++;
			colormap %= ncolormaps;
			break;
		case 'L':
			colormap += ncolormaps - 1;
			colormap %= ncolormaps;
			break;
		case 'j':
			plain_cosine = !plain_cosine;
			break;
		case 'z':
			fov /= 1.1f; camera.set_fov(fov); break;
		case 'Z':
			fov *= 1.1f; camera.set_fov(fov); break;
		case Ctrl+'x':
			printf("Writing %s\n", xffilename);
			xf.write(xffilename);
			break;
		case '.':
		case '>':
			if (only_band == 0)
				only_band = -1;
			else if (only_band < 0)
				only_band = 1;
			else if (only_band < data->sigmas.size() - 1)
				only_band++;
			else if (only_band == data->sigmas.size() - 1)
				only_band = 0;
			break;
		case ',':
		case '<':
			if (only_band == 0)
				only_band = data->sigmas.size() - 1;
			else if (only_band > 1)
				only_band--;
			else if (only_band == 1)
				only_band = -1;
			else if (only_band < 0)
				only_band = 0;
			break;
		case '!': preset(1); break;
		case '@': preset(2); break;
		case '#': preset(3); break;
		case '$': preset(4); break;
		case '\033': // Esc 
		case Ctrl+'q':
		case 'q':
			exit(0);
	}
	if (key == '0')
		only_band = (only_band == -1) ? 0 : -1;
	if (key >= '1' && key <= '9')
		only_band = key - '0';

	need_redraw();
	GLUI_Master.sync_live_all();
}

void skeyboardfunc(int key, int x, int y)
{
	switch (key) {
		case GLUT_KEY_UP:
			detail_contrast = min(max(detail_contrast * 1.1f,
						  detail_contrast + 0.01f),
					      2.0f);
			break;
		case GLUT_KEY_DOWN:
			detail_contrast = max(detail_contrast / 1.1f, 0.0f);
			break;
		case GLUT_KEY_RIGHT:
			base_contrast = min(max(base_contrast * 1.1f,
						base_contrast + 0.01f),
					    2.0f);
			break;
		case GLUT_KEY_LEFT:
			base_contrast = max(base_contrast / 1.1f, 0.0f);
			break;
	}
	need_redraw();
	GLUI_Master.sync_live_all();
}


// Reshape the window.  We clear the window here to possibly avoid some
// weird problems.  Yuck.
void reshape(int x, int y)
{
	GLUI_Master.auto_set_viewport();
	cls();
	glutSwapBuffers();
	need_redraw();
}


void usage(const char *myname)
{
	fprintf(stderr, "Usage: %s in.xsh\n", myname);
	exit(1);
}

int main(int argc, char *argv[])
{
	glutInitWindowSize(900, 750);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	glutInit(&argc, argv);

	if (argc < 2)
		usage(argv[0]);

	filename = argv[1];
	mesh = new TriMesh;
	data = new XSHdata;
	if (!xsh_read(filename, mesh, data))
		usage(argv[0]);

	xffilename = new char[strlen(filename) + 4];
	strcpy(xffilename, filename);
	char *dot = strrchr(xffilename, '.');
	if (!dot)
		dot = strrchr(xffilename, '\0');
	strcpy(dot, ".xf");

	const char *basename = strrchr(filename, '/');
	if (!basename)
		basename = strrchr(filename, '\\');
	if (!basename)
		basename = strrchr(filename, ':');
	if (!basename)
		basename = filename;
	else
		basename++;
	char *title = new char[strlen(basename) + 100];
	sprintf(title, "xshade: %s (%d vertices, %d smoothing levels)",
		basename, int(mesh->vertices.size()), int(data->sigmas.size())-1);
	int main_win = glutCreateWindow(title);
	glutDisplayFunc(redraw);
	GLUI_Master.set_glutMouseFunc(mousebuttonfunc);
	glutMotionFunc(mousemotionfunc);
	GLUI_Master.set_glutKeyboardFunc(keyboardfunc);
	GLUI_Master.set_glutSpecialFunc(skeyboardfunc); 
	GLUI_Master.set_glutReshapeFunc(reshape);

	GLUI *glui = GLUI_Master.create_glui_subwindow(main_win, GLUI_SUBWINDOW_BOTTOM);
	glui->set_main_gfx_window(main_win);
	GLUI_Rollout *g = glui->add_rollout("Options", false);


	glui->add_statictext_to_panel(g, "Base coat:");
	glui->add_slider_to_panel(g, "Contrast", GLUI_SLIDER_FLOAT,
				  0.0, 2.0, &base_contrast);
	glui->add_slider_to_panel(g, "log2 thresh", GLUI_SLIDER_FLOAT,
				  -8.0, 0.0, &log_base_thresh);

	glui->add_column_to_panel(g, false);

	glui->add_statictext_to_panel(g, "Exaggerated shading:");
	glui->add_slider_to_panel(g, "Contrast", GLUI_SLIDER_FLOAT,
				  0.0, 2.0, &detail_contrast);
	glui->add_slider_to_panel(g, "log2 thresh", GLUI_SLIDER_FLOAT,
				  -8.0, 0.0, &log_detail_thresh);
	//glui->add_button_to_panel(g, "Auto-set", 0, auto_detail_contrast);
	glui->add_slider_to_panel(g, "Scale dep", GLUI_SLIDER_FLOAT,
				  -1.0, 2.0, &detail_freqdep);
	//glui->add_slider_to_panel(g, "Skip # scales", GLUI_SLIDER_INT,
	//			  0, data->sigmas.size()-1, &skip_bands);

	glui->add_column_to_panel(g, false);

	glui->add_statictext_to_panel(g, "Mean curvature:");
	glui->add_slider_to_panel(g, "Contrast", GLUI_SLIDER_FLOAT,
				  0.0, 2.0, &curv_contrast);
	glui->add_slider_to_panel(g, "log2 thresh", GLUI_SLIDER_FLOAT,
				  -8.0, 0.0, &log_curv_thresh);
	glui->add_slider_to_panel(g, "Scale dep", GLUI_SLIDER_FLOAT,
				  -1.0, 2.0, &curv_freqdep);

	glui->add_column_to_panel(g, false);

	make_colormap_gui(glui, g);

	glui->add_column_to_panel(g, false);

	glui->add_statictext_to_panel(g, "Light direction:");
	glui->add_slider_to_panel(g, "Azimuth", GLUI_SLIDER_FLOAT,
				  -180.0, 180.0, &light_azim);
	glui->add_slider_to_panel(g, "Elevation", GLUI_SLIDER_FLOAT,
				  0, 80.0, &light_altitude);
	glui->add_slider_to_panel(g, "Azimuth 2", GLUI_SLIDER_FLOAT,
				  -180.0, 180.0, &light_azim2_rel);

	glui->add_column_to_panel(g, false);

	glui->add_checkbox_to_panel(g, "Just plain cosine", &plain_cosine);
	glui->add_checkbox_to_panel(g, "Use principal dir", &use_pdir);
	glui->add_statictext_to_panel(g, " ");
	glui->add_button_to_panel(g, "Preset 1 (general) ", 1, preset);
	glui->add_button_to_panel(g, "Preset 2 (terrain)  ", 2, preset);
	glui->add_button_to_panel(g, "Preset 3 (grooves)", 3, preset);
	glui->add_statictext_to_panel(g, " ");
	glui->add_button_to_panel(g, "Screencap", 0, dump_image);
	glui->add_button_to_panel(g, "Exit", 0, exit);

	resetview();
	preset(1);
	glutMainLoop();
}

